/* 
 * Integration Steeple 3 (USB ZigBee dongle) driver.
 *
 * Copyright 2005, 2006 Integration Associates Inc.
 * Written by Jon Beniston <jbeniston@integration.com>
 * 
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/module.h>
#include <linux/usb.h>

#include "steeple3.h"

/* Version Information */
#define DRIVER_VERSION 	"v0.3"
#define DRIVER_AUTHOR 	"Jon Beniston <jbeniston@integration.com>"
#define DRIVER_DESC 	"Integration Steeple 3 driver"

#define USB_STEEPLE3_INTEGRATION_VENDOR_ID		0x1795
#define USB_STEEPLE3_FRONTLINE_VENDOR_ID		0x16d3
#define USB_STEEPLE3_INTEGRATION_PRODUCT_ID		0x6002
#define USB_STEEPLE3_FRONTLINE_PRODUCT_ID		0xc002

static struct usb_device_id steeple3_table[] = {
	{USB_DEVICE(USB_STEEPLE3_INTEGRATION_VENDOR_ID, 0x6002)},
	{USB_DEVICE(USB_STEEPLE3_INTEGRATION_VENDOR_ID, 0x6004)},
	{USB_DEVICE(USB_STEEPLE3_INTEGRATION_VENDOR_ID, 0x6006)},
	{USB_DEVICE(USB_STEEPLE3_INTEGRATION_VENDOR_ID, 0x6008)},
	{USB_DEVICE(USB_STEEPLE3_INTEGRATION_VENDOR_ID, 0x600a)},
	{USB_DEVICE(USB_STEEPLE3_FRONTLINE_VENDOR_ID, 0xc002)},
	{} /* Terminating entry */
};
MODULE_DEVICE_TABLE(usb, steeple3_table);

/* FIXME: A proper number needs to be allocated */
#define USB_STEEPLE3_MINOR_BASE	208	

/* structure to hold all of our device specific stuff */
struct usb_steeple {
	struct usb_device *	udev;			/* the usb device pointer */
	struct usb_interface *	interface;		/* the interface for this device */
	char			num_bulk_in;		/* number of bulk in endpoints we have */
	char			num_bulk_out;		/* number of bulk out endpoints we have */

	unsigned char *		bulk_in_buffer;		/* the buffer to receive data */
	int			bulk_in_size;		/* the size of the receive buffer */
	__u8			bulk_in_endpointAddr;	/* the address of the bulk in endpoint */

	int 			rx_buffer_start;	/* index of where we currently are in buffer */
	int			rx_buffer_length;	/* how much data is in the buffer */

	unsigned char *		bulk_out_buffer;	/* the buffer to send data */
	int			bulk_out_size;		/* the size of the send buffer */
	__u8			bulk_out_endpointAddr;	/* the address of the bulk out endpoint */

        struct urb *            rx_urb;                 /* urb for reading data */
        wait_queue_head_t       queue;                  /* wait queue for read notification */
        struct semaphore        sem;                    /* */

	struct kref		kref;
};
#define to_steeple3_dev(d) container_of(d, struct usb_steeple, kref)

/* local function prototypes */
static void steeple3_urb_complete (struct urb *urb, struct pt_regs *regs);
static ssize_t steeple3_read	(struct file *file, char *buffer, size_t count, loff_t *ppos);
static ssize_t steeple3_write	(struct file *file, const char *buffer, size_t count, loff_t *ppos);
static unsigned int steeple3_poll (struct file *file, struct poll_table_struct *wait);
static int steeple3_ioctl	(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int steeple3_open	(struct inode *inode, struct file *file);
static int steeple3_release	(struct inode *inode, struct file *file);	
static int steeple3_probe	(struct usb_interface *interface, const struct usb_device_id *id);
static void steeple3_disconnect	(struct usb_interface *interface);

/* file operations needed when we register this driver. */
static struct file_operations steeple3_fops = {
	owner:		THIS_MODULE,
	read:		steeple3_read,
	write:		steeple3_write,
	poll:		steeple3_poll,
	ioctl:		steeple3_ioctl,
	open:		steeple3_open,
	release:	steeple3_release,
};      

/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver steeple3_driver = {
	owner: 		THIS_MODULE,
	name:		"steeple3",
	probe:		steeple3_probe,
	disconnect:	steeple3_disconnect,
	id_table:	steeple3_table,
};

/* usb class driver info in order to get a monor number from the usb core,
 * and to have the device registered with devfs and the driver core 
 */
static struct usb_class_driver steeple3_class = {
	name: "usb/steeple3%d",
	fops: &steeple3_fops,
	mode: S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH,
	minor_base: USB_STEEPLE3_MINOR_BASE,
};


/**
 *      steeple3_urb_complete 
 *
 *      Called when the read urb has completed.
 */
static void steeple3_urb_complete (struct urb *urb, struct pt_regs *regs)
{
        int retval;
        struct usb_steeple *dev = (struct usb_steeple *)urb->context;
        
        if (urb->status) {
                err ("Steeple 3: urb failed. status = %d", urb->status);       
                retval = usb_submit_urb (dev->rx_urb, GFP_KERNEL);      
                if (retval) {
                        err ("Steeple 3: Failed to resubmit urb. retval = %d", retval);
                }
        } else {
                /* update length of valid data in buffer */
		dev->rx_buffer_start = 0;
		dev->rx_buffer_length = 5 + dev->bulk_in_buffer[4];
                /* indicate that data has become available */          
                wake_up (&dev->queue);
        }
}


/**
 *	steeple3_delete
 */
static void steeple3_delete (struct kref *kref)
{
	struct usb_steeple *dev = to_steeple3_dev (kref);

	usb_put_dev (dev->udev);
	kfree (dev->bulk_in_buffer);
	kfree (dev);
}


/**
 *	steeple3_open
 */
static int steeple3_open (struct inode *inode, struct file *file)
{
	struct usb_steeple *dev;
	struct usb_interface *interface;
	int read;
	int subminor;
	int retval = 0;

	subminor = iminor(inode);

	interface = usb_find_interface (&steeple3_driver, subminor);
	if (!interface) {
		err ("Steeple 3: Can't find device for minor %d.", subminor);
		retval = -ENODEV;
		goto exit;
	}

	dev = usb_get_intfdata (interface);
	if (!dev) {
		retval = -ENODEV;
		goto exit;
	}

	/* increment usage count for the device */
	kref_get (&dev->kref);

	/* save our object in the file's private structure */
	file->private_data = dev;

        /* reset the cx1540 */
	retval = usb_control_msg(dev->udev, usb_sndctrlpipe(dev->udev, 0), 
				STEEPLE3_IOCTL_RESET, 
				USB_TYPE_VENDOR|USB_RECIP_ENDPOINT|USB_DIR_OUT,
				0, 0,
				NULL, 0, HZ);
	if (!retval) {
		/* power up the cx1540 */
		retval = usb_control_msg(dev->udev, usb_sndctrlpipe(dev->udev, 0), 
					STEEPLE3_IOCTL_WAKE, 
					USB_TYPE_VENDOR|USB_RECIP_ENDPOINT|USB_DIR_OUT,
					0, 0,
					NULL, 0, HZ);
		if (!retval) {
			/* read the MLME-SET.confirm that is generated when the 
			 * IEEE address is set after reset 
			 */
			do {
				retval = usb_bulk_msg (dev->udev,
						       usb_rcvbulkpipe (dev->udev, 
									dev->bulk_in_endpointAddr),
						       dev->bulk_in_buffer, dev->bulk_in_size,
						       &read, HZ*10);			
			} while ((retval > 0) && (dev->bulk_in_buffer[5] != 0x5c) && (dev->bulk_in_buffer[7] != 0x6f));
			if (retval < 0) {
				err ("Steeple 3: Failed to read MLME-SET.confirm (SUCCESS, macExtendedAddress). Devices extended address may not be set correctly. retval = %d.", retval);
			} 	
		} else {
			err ("Steeple 3: Failed to power up device. retval = %d.", retval);
		}
	} else {
		err ("Steeple 3: Failed to reset device. retval = %d.", retval);
	}

        /* initialise wait queue and semaphore */
        init_waitqueue_head (&dev->queue);
        sema_init (&dev->sem, 1);
	/* clear buffer */
	dev->rx_buffer_start = 0;
	dev->rx_buffer_length = 0;

        if (NULL != (dev->rx_urb = usb_alloc_urb (0, GFP_KERNEL))) {
                usb_fill_bulk_urb (dev->rx_urb, 
                                        dev->udev, 
                                        usb_rcvbulkpipe (dev->udev, dev->bulk_in_endpointAddr),
                                        dev->bulk_in_buffer, dev->bulk_in_size, 
                                        steeple3_urb_complete, dev);
                retval = usb_submit_urb (dev->rx_urb, GFP_KERNEL);
                if (retval) {
                        err ("Steeple 3: Failed to submit urb. retval = %d.", retval);
                }
        } else {
                retval = -ENOMEM;
        }
        

exit:
	return retval;
}


/**
 *	steeple3_release
 */
static int steeple3_release (struct inode *inode, struct file *file)
{
	struct usb_steeple *dev;

	dev = (struct usb_steeple *)file->private_data;
	if (dev == NULL) {
		return -ENODEV;
	}
	
	/* decrement usage count for the device */
	kref_put (&dev->kref, steeple3_delete);
	return 0;
}


/**
 *	steeple3_read
 */
static ssize_t steeple3_read (struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	struct usb_steeple *dev;
	int retval = 0;
	int copied = 0;
	int len;

        /* always just return 0, if count is 0 */
        if (count == 0) {
                return 0;
        }

        /* get pointer to our device specific data */
	dev = (struct usb_steeple *)file->private_data;

        /* enter critical section */	
        if (down_interruptible (&dev->sem)) 
                return -ERESTARTSYS;

        /* check to see if any data remains in our buffer */
        if (dev->rx_buffer_length == 0) {
                up (&dev->sem);
                if (file->f_flags & O_NONBLOCK) {        
                        return -EAGAIN;
                } else {
                        /* wait until data is available */
                        if (wait_event_interruptible (dev->queue, dev->rx_buffer_length > 0)) {
                                return -ERESTARTSYS;
                        } 
                        if (down_interruptible (&dev->sem)) 
                                return -ERESTARTSYS;
                }
        } 
        /* return as many bytes as possible upto the requested amount */
       	len = dev->rx_buffer_length - dev->rx_buffer_start;
	len = count > len ? len : count;
	if (len > 0) {
		if (copy_to_user (buffer, &dev->bulk_in_buffer[dev->rx_buffer_start], len)) {
                        up (&dev->sem);
			return -EFAULT;
		} else {
			copied = len;			
			dev->rx_buffer_start += len;
		}				
	}
        /* if buffer is now empty, resubmit urb to read some more data */
        if (dev->rx_buffer_length - dev->rx_buffer_start <= 0) {
                dev->rx_buffer_length = 0;
                dev->rx_buffer_start = 0;
                retval = usb_submit_urb (dev->rx_urb, GFP_KERNEL);
                if (retval) {
                        err ("Steeple 3: Failed to submit urb. retval = %d.", retval);
                        up (&dev->sem);
                        return retval;
                }
        }      

        /* exit critical section */
	up (&dev->sem);

        /* return the number of bytes that were read into the buffer */
	return copied;
}


/**
 *	steeple3_write
 */
static ssize_t steeple3_write (struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
	struct usb_steeple *dev;
	int bytes_written;
	int retval = 0;

	dev = (struct usb_steeple *)file->private_data;
				       
	/* verify that we actually have some data to write */
	if (count == 0) {
		goto exit;
	}
	if (count > dev->bulk_out_size) {
		retval = -EINVAL;
		goto exit;
	}

	/* copy the data from userspace into our buffer */
	if (copy_from_user(dev->bulk_out_buffer, buffer, 
			   count)) {
		retval = -EFAULT;
		goto exit;
	}

	/* do an blocking bulk write to send the data to the device */
	retval = usb_bulk_msg (dev->udev,
			       usb_sndbulkpipe (dev->udev, 
						dev->bulk_out_endpointAddr),
			       dev->bulk_out_buffer, count,
			       &bytes_written, HZ*10);
	if (!retval) {
		retval = bytes_written;
	}

exit:

	return retval;
}


/**
 *	steeple3_poll
 */
static unsigned int steeple3_poll (struct file *file, struct poll_table_struct *wait)
{
	struct usb_steeple *dev;
	unsigned int mask = 0;

	dev = (struct usb_steeple *)file->private_data;

        /* list which queues could indicate a change in status */
        poll_wait (file, &dev->queue, wait);
        /* indicate which operations can be performed without blocking */
	mask |= POLLOUT | POLLWRNORM;
	if (dev->rx_buffer_start < dev->rx_buffer_length) {
		mask |= POLLIN | POLLRDNORM;
	}

	return mask;
}


/**
 *	steeple3_ioctl
 */
static int steeple3_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct usb_steeple *dev;
	int retval = -ENOTTY;

	dev = (struct usb_steeple *)file->private_data;

	/* send the command to the device */
	if ((cmd >= STEEPLE3_IOCTL_RESET) && (cmd <= STEEPLE3_IOCTL_WAKE)) {
		retval = usb_control_msg(dev->udev, usb_sndctrlpipe(dev->udev, 0), 
				cmd, 
				USB_TYPE_VENDOR|USB_RECIP_ENDPOINT|USB_DIR_OUT,
				0, 0,
				NULL, 0, HZ);
	}
	
	return retval;
}

/**
 *	steeple3_probe
 *
 *	Called by the usb core when a new device is connected that it thinks
 *	this driver might be interested in.
 */
static int steeple3_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_steeple *dev = NULL;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	size_t buffer_size;
	int i;
	int retval = -ENOMEM;

	/* allocate memory for our device state and intialize it */
	dev = kmalloc (sizeof(*dev), GFP_KERNEL);
	if (dev == NULL) {
		err ("Steeple 3: %s:%d: Out of memory. Failed to allocate %d bytes.", __FUNCTION__, __LINE__, sizeof(*dev));
		goto error;
	}
	memset (dev, 0x00, sizeof (*dev));
	kref_init(&dev->kref);

	dev->udev = usb_get_dev(interface_to_usbdev(interface));
	dev->interface = interface;

	/* set up the endpoint information */
	iface_desc = interface->cur_altsetting;
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) 
			== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk in endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			dev->bulk_in_size = buffer_size;
			dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
			dev->bulk_in_buffer = kmalloc (buffer_size, GFP_KERNEL);
			if (!dev->bulk_in_buffer) {
				err("Steeple 3: %s:%d: Out of memory. Failed to allocate %d bytes.", __FUNCTION__, __LINE__, buffer_size);
				goto error;
			}
			dev->rx_buffer_start = 0;
			dev->rx_buffer_length = 0;
		}
		
		if (((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == 0x00) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) 
			== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk out endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			dev->bulk_out_size = buffer_size;
			dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
			dev->bulk_out_buffer = kmalloc (buffer_size, GFP_KERNEL);
			if (!dev->bulk_out_buffer) {
				err("Steeple 3: %s:%d: Out of memory. Failed to allocate %d bytes.", __FUNCTION__, __LINE__, buffer_size);
				goto error;
			}
		}
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, dev);
	
	/* register the device now, as it is ready */
	retval = usb_register_dev(interface, &steeple3_class);
	if (retval) {
		err("Steeple 3: Not able to get a minor for this device.");
		usb_set_intfdata(interface, NULL);
		goto error;
	}

	/* let the user know what node this device is now attached to */
	info ("USB steeple3%d connected.", interface->minor);
	return 0;
	
error:
	if (dev) {
		kref_put(&dev->kref, steeple3_delete);
	}
	return retval;
}


/**
 *	steeple3_disconnect
 *
 *	Called by the usb core when the device is removed from the system.
 */
static void steeple3_disconnect(struct usb_interface *interface)
{
	struct usb_steeple *dev;
	int minor = interface->minor;

	lock_kernel();

	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);

	usb_deregister_dev(interface, &steeple3_class);

	unlock_kernel();	

	kref_put(&dev->kref, steeple3_delete);

	info("USB steeple3%d disconnected.", minor);
}


/**
 *	usb_steeple3_init
 *
 *	Called when this module is loaded into the kernel.
 */
static int __init usb_steeple3_init(void)
{
	int result;

	/* register this driver with the USB subsystem */
	result = usb_register(&steeple3_driver);
	if (result != 0) {
		err("Steeple 3: %s:%d: usb_register failed. Error number %d",
		    __FUNCTION__, __LINE__, result);
	}
	return result;
}


/**
 *	usb_steeple3_exit
 *
 *	Called when this module is unloaded form the kernel.
 */
static void __exit usb_steeple3_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&steeple3_driver);
}


module_init (usb_steeple3_init);
module_exit (usb_steeple3_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
